/* Name: main.c
 * Project: Remote Sensor based on AVR USB driver
 * Author: Christian Starkjohann
 * Creation Date: 2005-03-20
 * Tabsize: 4
 * Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 * This Revision: $Id: sensord.c 316 2007-03-29 17:18:08Z cs $
 */

/*
General Description:
This program implements the host side of the "Remote Sensor". It receives data
via USB from the receiver circuit and logs it to a file in CSV format.
If you use this project for real data logging, you should check the calibration
constants which are hardcoded into this module (see function evaluate()).
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#ifndef __MINGW32_VERSION
#   include <syslog.h>
#   include <sys/errno.h>
#endif
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>
#include <time.h>
#include <math.h>
#include <usb.h>    /* this is libusb, see http://libusb.sourceforge.net/ */

#define USBDEV_SHARED_VENDOR    0x16C0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   0x05DC  /* Obdev's free shared PID */

#define USBDEV_VENDOR   0x16C0  /* VOTI */
#define USBDEV_PRODUCT  0x03eb  /* = 1003: Lab Test ID: Remote Sensor */
/* Our USB device identifies with the numbers above */

#define CMD_ECHO    0
#define CMD_POLL    1
#define CMD_READ    2
/* These are the vendor specific SETUP commands implemented by our USB device */

#ifdef __MINGW32_VERSION
#define sleep(x)    _sleep(x)
#endif

static char *filename = "/tmp/sensor.csv";
static FILE *fileFp = NULL;
static char debugMode = 0, didDetach = 0;

typedef struct calibration{
    double  tempOffset;
    double  tempScale;
    double  rhV50;
    double  rhV100;
    double  rhLin;
}calibration_t;

static calibration_t    calibration[16];

static void usage(char *name)
{
    fprintf(stderr, "usage: %s [-test] [-f csv-file] [-c calibration-file] [-v <vendor-id>] [-d <device-id>] [-debug]\n", name);
}

/* ------------------------------------------------------------------------- */

static void logVprintf(const char *format, va_list vlist)
{
#ifdef __MINGW32_VERSION
    vfprintf(stderr, format, vlist);
#else
static char didOpenLog = 0;

    if(debugMode || !didDetach){
        vfprintf(stderr, format, vlist);
    }
    if(didDetach){
        if(!didOpenLog){
            openlog("sensord", LOG_PID, LOG_DAEMON);
            didOpenLog = 1;
        }
        vsyslog(LOG_ERR, format, vlist);
    }
#endif
}

static void logPrintf(const char *format, ...)
{
va_list     vlist;

    va_start(vlist, format);
    logVprintf(format, vlist);
    va_end(vlist);
}

/* ------------------------------------------------------------------------- */

static void detachFromTerminal(void)
{
#ifndef __MINGW32_VERSION   /* Detaching only implemented for Unix */
int i;

/*  chdir("/"); do not chdir -- let the startup script decide */
/* do the double fork trick to detach reliably */
    switch(fork()){
        default:    _exit(0);   /* we are the parent: rval is child pid */
        case 0:     break;      /* child */
        case -1:                /* error */
            logPrintf("odSysDetach: error forking 1: [%d] %s\n", errno, strerror(errno));
            exit(1);
    }
    setpgid(getpid(), getpid());
    switch(fork()){
        default:    _exit(0);   /* we are the parent: rval is child pid */
        case 0:     break;      /* child */
        case -1:                /* error */
            logPrintf("odSysDetach: error forking 2: [%d] %s\n", errno, strerror(errno));
            exit(1);
    }
    /* we don't want random fds to become stdin/out/err, so force /dev/null for these fds: */
    for(i=0;i<3;i++)
        close(i);
    open("/dev/null", O_RDWR);
    dup(0);
    dup(0);
    umask(0);
#endif
    didDetach = 1;
}

/* ------------------------------------------------------------------------- */

static void initCalibration(void)
{
int i;

    for(i=0;i<16;i++){
        calibration[i].tempOffset = 0;  /* offset in Celsius */
        calibration[i].tempScale = 1;   /* applied after tempOffset */
        calibration[i].rhV50 = 0.3521;  /* sensor voltage at 50% rH */
        calibration[i].rhV100 = 0.63;   /* sensor voltage at 100% rH */
        calibration[i].rhLin = 50;      /* linearization factor (no lin -> infinite) */
    }
}

static void parseCalibrationFile(char *name)
{
FILE    *fp;
char    buffer[4096];
int     line = 0, sensor;
double  tempOffset, tempScale = 0, rhV50 = 0, rhV100 = 0, rhLin = 0;

    if((fp = fopen(name, "r")) == NULL){
        fprintf(stderr, "error opening calibration file \"%s\"\n", name);
        exit(1);
    }
    while(fgets(buffer, sizeof(buffer), fp) != NULL){
        line++;
        if(buffer[0] == '#')    /* remark */
            continue;
        if(buffer[0] == 0 || buffer[0] == '\n') /* empty line */
            continue;
        if(sscanf(buffer, "%d %lf %lf %lf %lf %lf", &sensor, &tempOffset, &tempScale, &rhV50, &rhV100, &rhLin) < 2){
            fprintf(stderr, "Error parsing line %d in file \"%s\":\n", line, name);
            fprintf(stderr, "  --->%s\n", buffer);
            exit(1);
        }
        if(sensor < 0 || sensor >= 16){
            fprintf(stderr, "Sensor number %d out of range in line %d of file \"%s\"\n", sensor, line, name);
            exit(1);
        }
        calibration[sensor].tempOffset = tempOffset;
        if(tempScale != 0)
            calibration[sensor].tempScale = tempScale;
        if(rhV50 != 0)
            calibration[sensor].rhV50 = rhV50;
        if(rhV100 != 0)
            calibration[sensor].rhV100 = rhV100;
        if(rhLin != 0)
            calibration[sensor].rhLin = rhLin;
    }
    fclose(fp);
}

static int  readInt(unsigned char *buffer)
{
    return *buffer + ((unsigned)buffer[1] << 8);
}


static void evaluate(unsigned char buffer[8])
{
char            timebuf[64];
struct tm       *tm;
time_t          now;
int             sequence = buffer[1] & 0xf, device = buffer[0];
int             iBatt = readInt(buffer+2), iTemp = readInt(buffer+4), iHumid = readInt(buffer+6);
static int      lastDevice = -1, lastSequence = -1;
static time_t   lastReceived;
double          vRef = 1.18;    /* band gap reference voltage */
double          vBatt, fTemp, fHumid, vHumid;

    device &= 0xf;  /* only 16 devices supported */
    time(&now);
    if(now - lastReceived < 5 && lastDevice == device && lastSequence == sequence)
        return; /* ignore retries */
    lastReceived = now;
    lastDevice = device;
    lastSequence = sequence;

    tm = localtime(&now);
    strftime(timebuf, sizeof(timebuf), "%Y-%m-%d %H:%M:%S", tm);
/* We may want to introduce device-specific calibration factors. But for now
 * use fixed compiled-in constants for all devices.
 */
    vBatt = 1024 * vRef / iBatt;    /* 1.18 V band gap reference on chip */
    fTemp = ((iTemp / 1.8 - 167.5) + calibration[device].tempOffset) * calibration[device].tempScale;   /* 1.8 counts / K */
    vHumid = vRef * iHumid / iBatt;
    if(vHumid > 1){ /* no diodes and no sensor present */
        fHumid = 0;
    }else{
        double a = 100 - 50 * calibration[device].rhV100 / calibration[device].rhV50;
        fHumid = 50/calibration[device].rhV50 * (vHumid + (0.043 - 0.0158 * vBatt));    /* battery voltage compensation */
        fHumid += a * exp(calibration[device].rhLin * (vHumid - calibration[device].rhV100));   /* series resistor compensation */
        if(fHumid > 100)
            fHumid = 100;   /* can't be more -- limit at 100 % */
    }
    if(fileFp != NULL){
        fprintf(fileFp, "%s;%3d;%6.2f;%6.1f;%4.0f;%7.3f\n", timebuf, device, vBatt, fTemp, fHumid, vHumid);
        fflush(fileFp);
    }
    if(debugMode){
        static char didPrint = 0;
        if(!didPrint){
            printf("%19s;%3s;%6s;%6s;%4s;%7s\n", "Date", "Dev", "Batt", "Temp", "Hum", "Raw Hum");
            didPrint = 1;
        }
        printf("%s;%3d;%6.2f;%6.1f;%4.0f;%7.3f\n", timebuf, device, vBatt, fTemp, fHumid, vHumid);
    }
}

/* ------------------------------------------------------------------------- */

static void signalReopenFile(int signalNr)
{
#ifndef __MINGW32_VERSION   /* MinGW does not emulate signals */
    signal(SIGHUP, signalReopenFile);   /* re-schedule for this signal */
#endif
    if(fileFp != NULL){
        fclose(fileFp);
        fileFp = NULL;
    }
    fileFp = fopen(filename, "a");
    if(fileFp == NULL){
        logPrintf("error opening file \"%s\" for append: [%d] %s\n", filename, errno, strerror(errno));
        exit(1);
    }
}

static char *getOptionArg(int *index, int argc, char **argv)
{
    (*index)++;
    if(*index >= argc){
        fprintf(stderr, "argument for option \"%s\" missing.\n", argv[*index - 1]);
        usage(argv[0]);
        exit(1);
    }
    return argv[*index];
}

static int  usbGetStringAscii(usb_dev_handle *dev, int index, int langid, char *buf, int buflen)
{
char    buffer[256];
int     rval, i;

    if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, buffer, sizeof(buffer), 1000)) < 0)
        return rval;
    if(buffer[1] != USB_DT_STRING)
        return 0;
    if((unsigned char)buffer[0] < rval)
        rval = (unsigned char)buffer[0];
    rval /= 2;
    /* lossy conversion to ISO Latin1 */
    for(i=1;i<rval;i++){
        if(i > buflen)  /* destination buffer overflow */
            break;
        buf[i-1] = buffer[2 * i];
        if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
            buf[i-1] = '?';
    }
    buf[i-1] = 0;
    return i-1;
}


#define USB_ERROR_NOTFOUND  1
#define USB_ERROR_ACCESS    2
#define USB_ERROR_IO        3

int usbOpenDevice(usb_dev_handle **device, int vendor, char *vendorName, int product, char *productName)
{
struct usb_bus      *bus;
struct usb_device   *dev;
usb_dev_handle      *handle = NULL;
int                 errorCode = USB_ERROR_NOTFOUND;
static int          didUsbInit = 0;

    if(!didUsbInit){
        didUsbInit = 1;
        usb_init();
    }
    usb_find_busses();
    usb_find_devices();
    for(bus=usb_get_busses(); bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next){
            if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product){
                char    string[256];
                int     len;
                handle = usb_open(dev); /* we need to open the device in order to query strings */
                if(!handle){
                    errorCode = USB_ERROR_ACCESS;
                    fprintf(stderr, "Warning: cannot open USB device: %s\n", usb_strerror());
                    continue;
                }
                if(vendorName == NULL && productName == NULL){  /* name does not matter */
                    break;
                }
                /* now check whether the names match: */
                len = usbGetStringAscii(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
                if(len < 0){
                    errorCode = USB_ERROR_IO;
                    fprintf(stderr, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
                }else{
                    errorCode = USB_ERROR_NOTFOUND;
                    /* fprintf(stderr, "seen device from vendor ->%s<-\n", string); */
                    if(strcmp(string, vendorName) == 0){
                        len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
                        if(len < 0){
                            errorCode = USB_ERROR_IO;
                            fprintf(stderr, "Warning: cannot query product for device: %s\n", usb_strerror());
                        }else{
                            errorCode = USB_ERROR_NOTFOUND;
                            /* fprintf(stderr, "seen product ->%s<-\n", string); */
                            if(strcmp(string, productName) == 0)
                                break;
                        }
                    }
                }
                usb_close(handle);
                handle = NULL;
            }
        }
        if(handle)
            break;
    }
    if(handle != NULL){
        errorCode = 0;
        *device = handle;
    }
    return errorCode;
}


int main(int argc, char **argv)
{
usb_dev_handle      *handle;
unsigned char       buffer[8];
char                runTest = 0;
int                 i, nBytes, vendor = USBDEV_VENDOR, device = USBDEV_PRODUCT;

    initCalibration();
    for(i=1;i<argc;i++){
        if(strcmp(argv[i], "-test") == 0){
            runTest = 1;
        }else if(strcmp(argv[i], "-f") == 0){
            filename = getOptionArg(&i, argc, argv);
        }else if(strcmp(argv[i], "-c") == 0){
            parseCalibrationFile(getOptionArg(&i, argc, argv));
        }else if(strcmp(argv[i], "-v") == 0){
            vendor = strtoul(getOptionArg(&i, argc, argv), 0, 0);
        }else if(strcmp(argv[i], "-d") == 0){
            device = strtoul(getOptionArg(&i, argc, argv), 0, 0);
        }else if(strcmp(argv[i], "-debug") == 0){
            debugMode = 1;
        }else{
            fprintf(stderr, "option \"%s\" not recognized.\n", argv[i]);
            usage(argv[0]);
            exit(1);
        }
    }
    if(!debugMode && !didDetach && !runTest)
        detachFromTerminal();
restart:    /* we jump back here if an error occurred */
#ifdef USE_SHARED_PID
    while(usbOpenDevice(&handle, USBDEV_SHARED_VENDOR, "www.obdev.at", USBDEV_SHARED_PRODUCT, "Remote Sensor") != 0){
#else
    while(usbOpenDevice(&handle, vendor, NULL, device, NULL) != 0){
#endif
        sleep(60);  /* wait until the device appears */
    }
    if(usb_set_configuration(handle, 1) < 0){
        logPrintf("error setting USB configuration: %s\n", usb_strerror());
    }
    if(usb_claim_interface(handle, 0) < 0){
        logPrintf("error setting USB interface: %s\n", usb_strerror());
    }
    if(runTest){
        int v, r;
/* The test consists of writing 1000 random numbers to the device and checking
 * the echo. This should discover systematic bit errors (e.g. in bit stuffing).
 */
        for(i=0;i<1000;i++){
            v = rand() & 0xffff;
            nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, CMD_ECHO, v, 0, (char *)buffer, sizeof(buffer), 5000);
            if(nBytes < 2){
                if(nBytes < 0)
                    logPrintf("USB error: %s\n", usb_strerror());
                logPrintf("only %d bytes received in iteration %d\n", nBytes, i);
                exit(1);
            }
            r = buffer[0] | (buffer[1] << 8);
            if(r != v){
                logPrintf("data error: received 0x%x instead of 0x%x in iteration %d\n", r, v, i);
                exit(1);
            }
        }
        logPrintf("test succeeded\n");
        exit(0);
    }
    signalReopenFile(1);    /* open file */
    for(;;){
        for(;;){    /* read as long as messages are available */
            nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, CMD_READ, 0, 0, (char *)buffer, sizeof(buffer), 5000);
            if(nBytes < 0){
                logPrintf("error in USB control transfer: %s\n", usb_strerror());
                goto usbErrorOccurred;
            }
            if(nBytes == 0) /* no more data */
                break;
            if(nBytes < 8){
                logPrintf("data format error, only %d bytes received (8 expected)\n", nBytes);
            }else{
                evaluate(buffer);
            }
        }
        /* wait for interrupt, set timeout to more than a week */
        nBytes = usb_interrupt_read(handle, USB_ENDPOINT_IN | 1 , (char *)buffer, sizeof(buffer), 700000 * 1000);
        if(nBytes < 0){
            logPrintf("error in USB interrupt read: %s\n", usb_strerror());
            goto usbErrorOccurred;
        }
    }
usbErrorOccurred:
    usb_close(handle);
    if(debugMode)
        exit(1);
    sleep(5);
    goto restart;
    return 0;
}

/* ------------------------------------------------------------------------- */
